@@ -19,8 +19,8 @@ TMPLPRO_LOCAL char*  pbuffer_init_as(pbuffer* pBuffer,size_t size);
 TMPLPRO_LOCAL char*  pbuffer_string(pbuffer*);
 TMPLPRO_LOCAL char*  pbuffer_resize(pbuffer*, size_t size);
 TMPLPRO_LOCAL void   pbuffer_free(pbuffer*);
-// currently unused
 TMPLPRO_LOCAL void pbuffer_fill_from_pstring(pbuffer* pBuffer, PSTRING pstr);
+TMPLPRO_LOCAL void pbuffer_swap(pbuffer* buf1, pbuffer* buf2);
 
 #endif /* pbuffer.h */
 
