@@ -15,17 +15,17 @@ typedef int boolval;
 
 struct tmplpro_state {
   boolval  is_visible;
-  char* top;
-  char* next_to_end;
-  char* last_processed_pos;
-  char* cur_pos;
+  const char* top;
+  const char* next_to_end;
+  const char* last_processed_pos;
+  const char* cur_pos;
   struct tmplpro_param* param;
   /* current tag */
   int   tag;
   boolval  is_tag_closed;
   boolval  is_tag_commented;
   boolval  is_expr;
-  char* tag_start; 
+  const char* tag_start; 
 
 /* internal buffers */
   /* main string buffer */
@@ -38,7 +38,7 @@ struct tmplpro_state {
   pbuffer expr_right_pbuffer;
 };
 
-extern TMPLPRO_LOCAL void _tmpl_log_state (struct tmplpro_state *state, int level);
+extern TMPLPRO_LOCAL void log_state(struct tmplpro_state*, int level, const char *fmt, ...) FORMAT_PRINTF(3,4);
 
 #endif /* prostate.h */
 
