@@ -18,8 +18,8 @@
 /*
  * generic load/first use library and unload/last use library hooks.
  */
-TMPLPRO_API void APICALL tmplpro_procore_init();
-TMPLPRO_API void APICALL tmplpro_procore_done();
+TMPLPRO_API void APICALL tmplpro_procore_init(void);
+TMPLPRO_API void APICALL tmplpro_procore_done(void);
 
 TMPLPRO_API const char* APICALL tmplpro_version(void);
 
@@ -29,28 +29,46 @@ struct tmplpro_param;
  * Note that struct tmplpro_param is not part of the interface
  * and is subject to change without notice.
  */
-TMPLPRO_API struct tmplpro_param* APICALL tmplpro_param_init();
+TMPLPRO_API struct tmplpro_param* APICALL tmplpro_param_init(void);
 TMPLPRO_API void APICALL tmplpro_param_free(struct tmplpro_param*);
 
 TMPLPRO_API int APICALL tmplpro_exec_tmpl (struct tmplpro_param*);
+TMPLPRO_API MPSTRING APICALL tmplpro_tmpl2pstring (struct tmplpro_param *param, int *exitcode);
+
+
+TMPLPRO_API void APICALL tmplpro_clear_option_param_map(struct tmplpro_param *param);
+TMPLPRO_API int APICALL tmplpro_count_option_param_map(struct tmplpro_param *param);
+TMPLPRO_API int APICALL tmplpro_push_option_param_map(struct tmplpro_param *param, ABSTRACT_MAP* map, EXPR_int64 flags);
 
 struct exprval;
 TMPLPRO_API void APICALL tmplpro_set_expr_as_int64 (ABSTRACT_EXPRVAL*,EXPR_int64);
 TMPLPRO_API void APICALL tmplpro_set_expr_as_double (ABSTRACT_EXPRVAL*,double);
-TMPLPRO_API void APICALL tmplpro_set_expr_as_string (ABSTRACT_EXPRVAL*,char*);
+TMPLPRO_API void APICALL tmplpro_set_expr_as_string (ABSTRACT_EXPRVAL*, const char*);
 TMPLPRO_API void APICALL tmplpro_set_expr_as_pstring (ABSTRACT_EXPRVAL*,PSTRING);
+TMPLPRO_API void APICALL tmplpro_set_expr_as_null (ABSTRACT_EXPRVAL*);
 
 TMPLPRO_API int  APICALL tmplpro_get_expr_type (ABSTRACT_EXPRVAL*);
 TMPLPRO_API EXPR_int64 APICALL tmplpro_get_expr_as_int64 (ABSTRACT_EXPRVAL*);
 TMPLPRO_API double APICALL tmplpro_get_expr_as_double (ABSTRACT_EXPRVAL*);
-TMPLPRO_API char* APICALL tmplpro_get_expr_as_string (ABSTRACT_EXPRVAL*);
 TMPLPRO_API PSTRING APICALL tmplpro_get_expr_as_pstring (ABSTRACT_EXPRVAL*);
 
+
+#define ASK_NAME_DEFAULT	0
+#define ASK_NAME_AS_IS		1
+#define ASK_NAME_LOWERCASE	2
+#define ASK_NAME_UPPERCASE	4
+#define ASK_NAME_MASK	(ASK_NAME_AS_IS|ASK_NAME_LOWERCASE|ASK_NAME_UPPERCASE)
+/* future compatibility: not yet implemented */
+#define ASK_NAME_CAPITALIZED	8
+#define ASK_NAME_LCFIRST	16
+#define ASK_NAME_UCFIRST	32
+/* define ASK_NAME_MASK	(ASK_NAME_AS_IS|ASK_NAME_LOWERCASE|ASK_NAME_UPPERCASE|ASK_NAME_CAPITALIZED|ASK_NAME_LCFIRST|ASK_NAME_UCFIRST) */
+
 #endif /* tmplpro.h */
 
 
 
-/*! \fn void tmplpro_procore_init();
+/*! \fn void tmplpro_procore_init(void);
     \brief generic load library/first use initializer.
 
     Initializer of global internal structures.
@@ -59,7 +77,7 @@ TMPLPRO_API PSTRING APICALL tmplpro_get_expr_as_pstring (ABSTRACT_EXPRVAL*);
     \warning May not be thread safe. Should be called once.
 */
 
-/*! \fn void tmplpro_procore_done();
+/*! \fn void tmplpro_procore_done(void);
     \brief generic load/first use library and unload/last use library hooks.
 
     Deinitializer of global internal structures.
@@ -73,7 +91,7 @@ TMPLPRO_API PSTRING APICALL tmplpro_get_expr_as_pstring (ABSTRACT_EXPRVAL*);
     \return version string.
 */
 
-/*! \fn struct tmplpro_param* tmplpro_param_init();
+/*! \fn struct tmplpro_param* tmplpro_param_init(void);
     \brief Constructor of tmplpro_param.
 */
 
@@ -85,6 +103,12 @@ TMPLPRO_API PSTRING APICALL tmplpro_get_expr_as_pstring (ABSTRACT_EXPRVAL*);
     \brief main method of libhtmltmplpro.
 */
 
+/*! \fn MPSTRING tmplpro_tmpl2pstring (struct tmplpro_param*, int* exitcode);
+    \brief main method of libhtmltmplpro. Returns processed template as a C string.
+
+    Note that caller's responsibility is to free the returned memory.
+*/
+
 /*! \fn void tmplpro_set_expr_as_int64 (ABSTRACT_EXPRVAL*,EXPR_int64);
     \brief method to return int64 value from callback of call_expr_userfnc_functype.
 
@@ -109,6 +133,12 @@ TMPLPRO_API PSTRING APICALL tmplpro_get_expr_as_pstring (ABSTRACT_EXPRVAL*);
     It should only be used in a callback of call_expr_userfnc_functype.
 */
 
+/*! \fn void tmplpro_set_expr_as_null (ABSTRACT_EXPRVAL*);
+    \brief method to return null from callback of call_expr_userfnc_functype.
+
+    It should only be used in a callback of call_expr_userfnc_functype.
+*/
+
 /*! \fn EXPR_int64 tmplpro_get_expr_as_int64 (ABSTRACT_EXPRVAL*);
     \brief method for callback of push_expr_arglist_functype to retrieve a value as int64.
 */
@@ -119,12 +149,6 @@ TMPLPRO_API PSTRING APICALL tmplpro_get_expr_as_pstring (ABSTRACT_EXPRVAL*);
     It should only be used in a callback of push_expr_arglist_functype.
 */
 
-/*! \fn char* tmplpro_get_expr_as_string (ABSTRACT_EXPRVAL*);
-    \brief method for callback of push_expr_arglist_functype to retrieve a value as C string.
-
-    It should only be used in a callback of push_expr_arglist_functype.
-*/
-
 /*! \fn PSTRING tmplpro_get_expr_as_pstring (ABSTRACT_EXPRVAL*);
     \brief method for callback of push_expr_arglist_functype to retrieve a value as PSTRING.
 
@@ -148,7 +172,7 @@ TMPLPRO_API PSTRING APICALL tmplpro_get_expr_as_pstring (ABSTRACT_EXPRVAL*);
     
     Destructor is tmplpro_param_free()
 
-    Main method is tmplpro_exec_tmpl()
+    Main methods are tmplpro_exec_tmpl(), tmplpro_tmpl2pstring()
 
  */
 
@@ -161,15 +185,15 @@ TMPLPRO_API PSTRING APICALL tmplpro_get_expr_as_pstring (ABSTRACT_EXPRVAL*);
     It can contain string, 64-bit integer or double.
 
     Methods:
+    \li tmplpro_set_expr_as_null(ABSTRACT_EXPRVAL*)
     \li tmplpro_set_expr_as_int64(ABSTRACT_EXPRVAL*,EXPR_int64)
     \li tmplpro_set_expr_as_double(ABSTRACT_EXPRVAL*,double)
-    \li tmplpro_set_expr_as_string(ABSTRACT_EXPRVAL*,char*)
+    \li tmplpro_set_expr_as_string(ABSTRACT_EXPRVAL*,const char*)
     \li tmplpro_set_expr_as_pstring(ABSTRACT_EXPRVAL*,PSTRING)
 
     \li tmplpro_get_expr_type(ABSTRACT_EXPRVAL*)
     \li tmplpro_get_expr_as_int64(ABSTRACT_EXPRVAL*)
     \li tmplpro_get_expr_as_double(ABSTRACT_EXPRVAL*)
-    \li tmplpro_get_expr_as_string(ABSTRACT_EXPRVAL*)
     \li tmplpro_get_expr_as_pstring(ABSTRACT_EXPRVAL*)
  */
 
